from pydantic import BaseModel, ConfigDict, Field, alias_generators

from typing import List, Literal, Optional, Union

import uuid6 as uuid
from enum import Enum


def uuidv7():
    return str(uuid.uuid7())


class Base(BaseModel):
    model_config = ConfigDict(
        populate_by_name=True,
        alias_generator=alias_generators.to_camel,
        validate_assignment=True,
        protected_namespaces=(),
    )


# In the sandbox, all resources that can directly be accessed by
# users are logically scoped by intermediaries. This is a base class
# for such resources.
class ScopedByIntermediaryBase(Base):
    intermediary_id: str = Field(alias='intermediaryId')


# Enums / Constants


class AccountKind(str, Enum):
    ORDINARY = 'Ordinary'
    CORPORATION = 'Corporation'


class AccountState(str, Enum):
    ACTIVE = 'Active'
    INACTIVE = 'Inactive'
    FREEZED = 'Freezed'


class BalanceKind(str, Enum):
    ORDINARY = 'Ordinary'


class BalanceState(str, Enum):
    ACTIVE = 'Active'
    INACTIVE = 'Inactive'


class BalanceConnectorKind(str, Enum):
    HOLDING = 'Holding'
    HTLC_SENDING = 'HTLCSending'
    HTLC_RECEIVING = 'HTLCReceiving'


class TransactionKind(str, Enum):
    TRANSFER = 'Transfer'
    FUNDING = 'Funding'
    DEFUNDING = 'Defunding'


class BalanceUpdateType(str, Enum):
    INCOMING = 'Incoming'
    OUTGOING = 'Outgoing'


class BalanceSelectorType(str, Enum):
    BALANCE = 'Balance'
    CONNECTOR = 'Connector'


class BalanceSpecifier(Base):
    type: Literal[BalanceSelectorType.BALANCE]
    id: str


class ConnectorSpecifier(Base):
    type: Literal[BalanceSelectorType.CONNECTOR]
    id: str


MAX_BALANCE: int = 9223372036854775807


# Models


class Intermediary(Base):
    id: str = Field(default_factory=uuidv7)
    name: str


class Balance(ScopedByIntermediaryBase):
    id: str = Field(default_factory=uuidv7)
    state: str = Field(default=BalanceState.ACTIVE)
    kind: str = Field(default=BalanceKind.ORDINARY)
    amount: int = Field(default=0)


class BalanceConnector(ScopedByIntermediaryBase):
    id: str = Field(default_factory=uuidv7)
    account_id: str = Field(alias='accountId')
    balance_id: str = Field(alias='balanceId')
    kind: str
    permission: List[str] = Field(default=['Viewable', 'Spendable', 'Receivable'])


class NewConnectorOnBalance(ScopedByIntermediaryBase):
    id: str = Field(default_factory=uuidv7)
    account_id: str = Field(alias='accountId')
    kind: str
    permission: List[str] = Field(default=['Viewable', 'Spendable', 'Receivable'])


class BalanceWithConnectors(Balance):
    connectors: List[NewConnectorOnBalance]


class Account(ScopedByIntermediaryBase):
    id: str = Field(default_factory=uuidv7)
    holder_name: str = Field(alias='holderName')
    state: str = Field(default=AccountState.ACTIVE)
    kind: str = Field(default=AccountKind.ORDINARY)
    balance: int = Field(default=0)


class HTLCConnector(BalanceConnector):
    contract_id: str = Field(alias='contractId')


class HTLCContract(Base):
    id: str = Field(default_factory=uuidv7)
    hash: str
    expires_at: str


class TransactionBase(ScopedByIntermediaryBase):
    # Overriding ScopedByIntermediaryBase.intermediary_id.
    # This should be filled by referencing account's intermediary.
    intermediary_id: Optional[str] = \
        Field(alias='intermediaryId', default=None)

    id: str = Field(default_factory=uuidv7)
    sub_id: int = Field(default=0)
    datetime: Optional[int] = Field(default=None)
    kind: str
    attachment: Optional[str] = Field(default=None)


class BalanceUpdate(ScopedByIntermediaryBase):
    # Overriding ScopedByIntermediaryBase.intermediary_id.
    # This should be filled by referencing account's intermediary.
    intermediary_id: Optional[str] = \
        Field(alias='intermediaryId', default=None)

    transaction_id: Optional[str] = Field(alias='transactionId', default=None)
    connector_id: str = Field(alias='balanceConnectorId')
    made_for: Optional[str] = Field(alias='madeFor', default=None)
    amount: int
    type: str
    balance_after_transaction: Optional[int] = \
        Field(alias='balanceAfterTransaction', default=None)


class CustomTransaction(TransactionBase):
    # Overriding TransactionBase.intermediary_id.
    intermediary_id: str = Field(alias='intermediaryId')

    balance_updates: List[BalanceUpdate]


# Transaction variations.


class Payment(TransactionBase):
    to: Union[str, ConnectorSpecifier]
    source_balance_selector: Optional[Union[BalanceSpecifier, ConnectorSpecifier]]\
        = Field(alias='sourceBalanceSelector', default=None)
    amount: int
    # Overriding Transaction.kind.
    kind: Literal[TransactionKind.TRANSFER] =\
        Field(default=TransactionKind.TRANSFER)


class Funding(TransactionBase):
    amount: int
    # Overriding Transaction.kind.
    kind: Literal[TransactionKind.FUNDING] =\
        Field(default=TransactionKind.FUNDING)
    balance_selector: Optional[Union[BalanceSpecifier, ConnectorSpecifier]]\
        = Field(alias='balanceSelector', default=None)


class Defunding(TransactionBase):
    amount: int
    # Overriding Transaction.kind.
    kind: Literal[TransactionKind.DEFUNDING] =\
        Field(default=TransactionKind.DEFUNDING)
    balance_selector: Optional[Union[BalanceSpecifier, ConnectorSpecifier]]\
        = Field(alias='balanceSelector', default=None)


class HTLC(Base):
    to: str
    amount: int
    hash: str
    expires_at: str = Field(alias='expiresAt')


class HTLCSecret(Base):
    secret: bytes


# For patch


class AccountUpdate(Base):
    holder_name: Optional[str] = Field(alias='holderName', default=None)
    state: Optional[str] = Field(default=None)


class BalanceStateUpdate(Base):
    state: Optional[str] = Field(default=None)
    kind: Optional[str] = Field(default=None)


# Others


class Preset(Base):
    version: str
    intermediaries: List[Intermediary]
