import datetime
import base64
import hashlib

from sqlalchemy import func, orm, select, delete
from sqlalchemy.exc import NoResultFound, IntegrityError
from sqlalchemy import sql

from fastapi import HTTPException

import api_schemas as api
import db_models as db

from typing import Optional


# This code makes frequent use of JOINs, which may cause performance
# issues. For the time being, however, we trust the power of index and
# DB's query optimizer, and will not perform manual optimization until
# the problem appears.


# Intermediary


def read_intermediaries(session: orm.Session):
    return [intm.to_dict() for intm in session.scalars(select(db.Intermediary)).all()]


# Account


def create_account(session: orm.Session, new_account: api.Account):
    new_balance = api.Balance(
        intermediary_id=new_account.intermediary_id
    )
    new_connector = api.BalanceConnector(
        intermediary_id=new_account.intermediary_id,
        account_id=new_account.id,
        balance_id=new_balance.id,
        kind=api.BalanceConnectorKind.HOLDING,
    )

    account = db.Account(
        id=new_account.id,
        intermediary_id=new_account.intermediary_id,
        holder_name=new_account.holder_name,
        state=new_account.state,
        kind=new_account.kind,
        note=None,
        connectors=[
            db.BalanceConnector(
                id=new_connector.id,
                intermediary_id=new_connector.intermediary_id,
                account_id=new_connector.account_id,
                balance_id=new_connector.balance_id,
                kind=new_connector.kind,
                permission=db.Permission.from_list(
                    new_connector.permission
                ).value,
                balance=db.Balance(
                    id=new_balance.id,
                    state=new_balance.state,
                    kind=new_balance.kind,
                    amount=new_balance.amount,
                    intermediary_id=new_balance.intermediary_id,
                )
            ),
        ]
    )

    try:
        session.add(account)
        session.commit()
    except IntegrityError:
        raise HTTPException(
            status_code=400, detail='Invalid intermediary specified'
        )

    return account.to_dict()


def update_account(
        session: orm.Session,
        account_id: str,
        update: api.AccountUpdate,
):
    stmt = select(db.Account).where(db.Account.id == account_id)

    target = session.scalars(stmt).one()

    if update.holder_name:
        target.holder_name = update.holder_name
    if update.state:
        target.state = update.state

    try:
        session.commit()

        return target.to_dict()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such account')


def read_accounts(session: orm.Session, intermediary_id: str):
    stmt = select(db.Account)\
        .where(db.Account.intermediary_id == intermediary_id)

    accounts = session.scalars(stmt).all()

    return [acc.to_dict() for acc in accounts]


def read_account_by_id(session: orm.Session, account_id: str):
    stmt = select(db.Account).where(db.Account.id == account_id)

    try:
        return session.scalars(stmt).one().to_dict()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such account')


def _check_if_account_exists(session: orm.Session, account_id: str):
    _ = read_account_by_id(session, account_id)


def _check_account(session: orm.Session, account_id: str):
    acc = read_account_by_id(session, account_id)

    if acc['state'] == api.AccountState.FREEZED:
        raise HTTPException(status_code=403, detail='Account is freezed')
    if acc['state'] == api.AccountState.INACTIVE:
        raise HTTPException(status_code=403, detail='Account is inactive')


def read_accounts_with_balance(session: orm.Session, intermediary_id: str):
    # Use outer join to retrieve account even when account has no connected balance.
    stmt = select(db.Account, func.sum(db.Balance.amount).label('balance'))\
        .outerjoin(db.Account.connectors)\
        .outerjoin(db.BalanceConnector.balance)\
        .where(db.Account.intermediary_id == intermediary_id)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.VIEWABLE.value))\
        .group_by(db.Account.id)

    # To get both of account and its balance, use execute() instead of scalars().
    accounts = session.execute(stmt).all()

    # Note: If an account has no connectors/balances, balance is None.
    return [{**acc.to_dict(), 'balance': int(bal or 0)} for (acc, bal) in accounts]


def read_account_by_id_with_balance(session: orm.Session, account_id: str):
    # Use outer join to retrieve account even when account has no connected balance.
    stmt = select(db.Account, func.sum(db.Balance.amount).label('balance'))\
        .outerjoin(db.Account.connectors)\
        .outerjoin(db.BalanceConnector.balance)\
        .where(db.Account.id == account_id)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.VIEWABLE.value))\
        .group_by(db.Account.id)

    # To get both of account and its balance, use execute() instead of scalars().
    try:
        account, balance = session.execute(stmt).one()

        # Note: If an account has no connectors/balances, balance is None.
        return {**account.to_dict(), 'balance': int(balance or 0)}
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such account')


def delete_account_by_id(session: orm.Session, account_id: str):
    stmt = select(db.Account)\
        .join(db.Account.connectors)\
        .join(db.BalanceConnector.balance)\
        .where(db.Account.id == account_id)\
        .distinct()

    try:
        account = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such account')

    # Delete balances connected to the account if it is not connected to
    # other accounts.

    is_other_account_connected = select(db.BalanceConnector)\
        .where(db.BalanceConnector.balance_id == db.Balance.id)\
        .where(db.BalanceConnector.account_id != account_id)\
        .exists()\
        .correlate(db.Balance)

    stmt = delete(db.Balance)\
        .where(db.Balance.id == db.BalanceConnector.balance_id)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(~is_other_account_connected)

    session.execute(stmt)

    session.execute(
        delete(db.Account).where(db.Account.id == account_id)
    )

    session.commit()


# Balance


def create_balance(session: orm.Session, new_balance: api.BalanceWithConnectors):
    balance = db.Balance(
        id=new_balance.id,
        state=new_balance.state,
        kind=new_balance.kind,
        amount=new_balance.amount,
        intermediary_id=new_balance.intermediary_id,
        connectors=[
            db.BalanceConnector(
                id=new_connector.id,
                account_id=new_connector.account_id,
                balance_id=new_balance.id,
                kind=new_connector.kind,
                intermediary_id=new_connector.intermediary_id,
                permission=db.Permission.from_list(
                    new_connector.permission
                ).value,
            ) for new_connector in new_balance.connectors
        ]
    )

    try:
        session.add(balance)
        session.commit()
    except IntegrityError:
        raise HTTPException(
            status_code=400, detail='Invalid intermediary specified'
        )

    return balance.to_dict()


def read_balances(session: orm.Session, intermediary_id: str):
    stmt = select(db.Balance)\
        .where(db.Balance.intermediary_id == intermediary_id)

    balances = session.scalars(stmt).all()

    return [bal.to_dict() for bal in balances]


def read_balances_connected_to_account(session: orm.Session, account_id: str):
    stmt = select(db.Account)\
        .outerjoin(db.Account.connectors)\
        .outerjoin(db.BalanceConnector.balance)\
        .where(db.Account.id == account_id)\
        .distinct()

    try:
        account = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such account')

    return [
        conn.balance.to_dict_with_connector(conn) for conn in account.connectors
    ]


def read_balance_by_id(session: orm.Session, balance_id: str):
    stmt = select(db.Balance).where(db.Balance.id == balance_id)

    try:
        return session.scalars(stmt).one().to_dict()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such balance')


def read_balance_connected_to_account_by_id(
        session: orm.Session,
        account_id: str,
        balance_id: str,
):
    _check_if_account_exists(session, account_id)

    stmt = select(db.BalanceConnector)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.balance_id == balance_id)

    try:
        connector = session.scalars(stmt).one()

        return connector.balance.to_dict_with_connector(connector)
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such balance')


def update_balance(
        session: orm.Session,
        balance_id: str,
        update: api.BalanceStateUpdate
):
    stmt = select(db.Balance).where(db.Balance.id == balance_id)

    target = session.scalars(stmt).one()

    if update.kind:
        target.kind = update.kind
    if update.state:
        target.state = update.state

    try:
        session.commit()

        return target.to_dict()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such balance')


def delete_balance_by_id(session: orm.Session, balance_id: str):
    result = session.execute(
        delete(db.Balance).where(db.Balance.id == balance_id)
    )

    if result.rowcount == 0:
        raise HTTPException(status_code=404, detail='No such balance')

    session.commit()


# Connectors


def create_connector(session: orm.Session, new_connector: api.BalanceConnector):
    connector = db.BalanceConnector(
        id=new_connector.id,
        account_id=new_connector.account_id,
        balance_id=new_connector.balance_id,
        intermediary_id=new_connector.intermediary_id,
        kind=new_connector.kind,
        permission=db.Permission.from_list(new_connector.permission).value,
    )

    try:
        session.add(connector)
        session.commit()
    except IntegrityError:
        raise HTTPException(
            status_code=400,
            detail='Invalid intermediary or account or balance specified',
        )

    return connector.to_dict()


def read_connectors(
        session: orm.Session,
        intermediary_id: str,
        **filters,
):
    stmt = select(db.BalanceConnector)\
        .where(db.Intermediary.id == intermediary_id)

    if 'account_id' in filters and filters['account_id']:
        stmt = \
            stmt.where(db.BalanceConnector.account_id == filters['account_id'])
    if 'balance_id' in filters and filters['balance_id']:
        stmt = \
            stmt.where(db.BalanceConnector.balance_id == filters['balance_id'])
    if 'kind' in filters and filters['kind']:
        stmt = stmt.where(db.BalanceConnector.kind == filters['kind'])

    connectors = session.scalars(stmt).all()

    return [conn.to_dict() for conn in connectors]


def read_connector_by_id(session: orm.Session, connector_id: str):
    stmt = select(db.BalanceConnector)\
        .where(db.BalanceConnector.id == connector_id)

    try:
        return session.scalars(stmt).one().to_dict()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such connector')


def delete_connector_by_id(session: orm.Session, connector_id: str):
    stmt = delete(db.BalanceConnector)\
        .where(db.BalanceConnector.id == connector_id)

    result = session.execute(stmt)

    if result.rowcount == 0:
        raise HTTPException(status_code=404, detail='No such connector')
    session.commit()


# HTLC connector


def read_htlc_connectors(session: orm.Session, account_id: str):
    _check_if_account_exists(session, account_id)

    stmt = select(db.HTLCConnector)\
        .join(db.HTLCConnector.contract)\
        .join(db.HTLCConnector.inner)\
        .where(db.BalanceConnector.account_id == account_id)\
        .distinct()

    connectors = session.scalars(stmt).all()

    return [conn.to_dict() for conn in connectors]


def read_htlc_connector_by_id(session: orm.Session, connector_id: str):
    stmt = select(db.HTLCConnector)\
        .join(db.HTLCConnector.contract)\
        .join(db.HTLCConnector.inner)\
        .where(db.BalanceConnector.id == connector_id)\
        .distinct()

    try:
        return session.scalars(stmt).one().to_dict()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such connector')


def delete_htlc_connector_with_retake(
        session: orm.Session,
        account_id: str,
        connector_id: str,
):
    _check_account(session, account_id)

    stmt = select(db.HTLCConnector)\
        .join(db.HTLCConnector.contract)\
        .join(db.HTLCConnector.inner)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.kind == api.BalanceConnectorKind.HTLC_SENDING)\
        .where(db.HTLCConnector.id == connector_id)\
        .distinct()

    try:
        connector = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such connector')

    stmt = select(db.BalanceConnector)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.RECEIVABLE.value) > 0)\
        .limit(1)\
        .distinct()

    try:
        sender = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(
            status_code=422, detail='Corresponding connector not found')

    new_transaction = api.TransactionBase(kind=api.TransactionKind.TRANSFER)

    # Validate the lock condition.
    if datetime.datetime.now() < connector.contract.expires_at:
        raise HTTPException(status_code=422, detail='Lock not expired')

    # Retake the locked amount.

    if sender.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")

    if sender.balance.amount + connector.inner.balance.amount > api.MAX_BALANCE:
        raise HTTPException(status_code=400, detail='Too large amount')

    sender.balance.amount += connector.inner.balance.amount
    # Clean up the locked balance and its all connectors.
    stmt = sql.text('''
    DELETE balance_connector, htlc_connector, htlc, balance
      FROM balance_connector
           INNER JOIN htlc_connector ON htlc_connector.id = balance_connector.id
           INNER JOIN htlc ON htlc.id = htlc_connector.contract_id
           INNER JOIN balance ON balance_connector.balance_id = balance.id
    WHERE htlc_connector.contract_id = :contract_id;
    ''')

    session.execute(stmt, {'contract_id': connector.contract_id})

    # Record transaction.
    transaction = db.Transaction(
        id=new_transaction.id,
        kind=new_transaction.kind,
        note=new_transaction.attachment,
        intermediary_id=sender.intermediary_id,
        balance_updates=[
            db.BalanceUpdate(
                balance_id=connector.inner.balance_id,
                amount=connector.inner.balance.amount,
                type=api.BalanceUpdateType.OUTGOING,
                balance_after_transaction=0,
                intermediary_id=sender.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=connector.inner.balance_id
                    ) for conn in connector.inner.balance.connectors if conn.is_viewable()
                ],
            ),
            db.BalanceUpdate(
                balance_id=sender.balance.id,
                amount=connector.inner.balance.amount,
                type=api.BalanceUpdateType.INCOMING,
                balance_after_transaction=sender.balance.amount,
                intermediary_id=sender.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=sender.balance.id
                    ) for conn in sender.balance.connectors if conn.is_viewable()
                ],
            ),
        ]
    )
    session.add(transaction)
    session.commit()

    return transaction.to_dict(filter_for=account_id)


def delete_htlc_connector_with_claim(
        session: orm.Session,
        account_id: str,
        connector_id: str,
        secret: api.HTLCSecret,
):
    _check_account(session, account_id)

    stmt = select(db.HTLCConnector)\
        .join(db.HTLCConnector.contract)\
        .join(db.HTLCConnector.inner)\
        .join(db.BalanceConnector.balance)\
        .where(db.HTLCConnector.id == connector_id)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.kind == api.BalanceConnectorKind.HTLC_RECEIVING)\
        .distinct()

    try:
        connector = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such connector')

    stmt = select(db.BalanceConnector)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.RECEIVABLE.value) > 0)\
        .limit(1)\
        .distinct()

    try:
        receiver = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(
            status_code=422, detail="Receiver's available balance not found"
        )

    stmt = select(db.HTLCConnector)\
        .join(db.HTLCConnector.inner)\
        .where(db.HTLCConnector.contract_id == connector.contract_id)\
        .where(db.BalanceConnector.kind == api.BalanceConnectorKind.HTLC_SENDING)

    try:
        sender = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(
            status_code=422, detail="Sender's connector not found"
        )

    new_transaction = api.TransactionBase(kind=api.TransactionKind.TRANSFER)

    # Validate the condition.

    if datetime.datetime.now() > connector.contract.expires_at:
        raise HTTPException(status_code=400, detail='Lock expired')

    try:
        hash = hashlib.sha256(base64.b64decode(secret.secret)).hexdigest()
    except TypeError as e:
        raise HTTPException(
            status_code=400, detail='`secret` must be baes64-encoded'
        )

    if hash != connector.contract.hash:
        raise HTTPException(status_code=400, detail='Hashes not match')

    # Move the locked amount.

    if receiver.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")

    if receiver.balance.amount + connector.inner.balance.amount > api.MAX_BALANCE:
        raise HTTPException(status_code=400, detail='Too large amount')

    receiver.balance.amount += connector.inner.balance.amount
    # Clean up the locked balance and its connectors.
    stmt = sql.text('''
    DELETE balance_connector, htlc_connector, htlc, balance
      FROM balance_connector
           INNER JOIN htlc_connector ON htlc_connector.id = balance_connector.id
           INNER JOIN htlc ON htlc.id = htlc_connector.contract_id
           INNER JOIN balance ON balance_connector.balance_id = balance.id
    WHERE htlc_connector.contract_id = :contract_id;
    ''')

    # Record transaction.
    transaction = db.Transaction(
        id=new_transaction.id,
        kind=new_transaction.kind,
        note=new_transaction.attachment,
        intermediary_id=receiver.intermediary_id,
        balance_updates=[
            db.BalanceUpdate(
                balance_id=connector.inner.balance_id,
                amount=connector.inner.balance.amount,
                type=api.BalanceUpdateType.OUTGOING,
                balance_after_transaction=0,
                intermediary_id=sender.inner.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=connector.inner.balance_id,
                    ) for conn in connector.inner.balance.connectors if conn.is_viewable()
                ],
            ),
            db.BalanceUpdate(
                balance_id=receiver.balance.id,
                amount=connector.inner.balance.amount,
                type=api.BalanceUpdateType.INCOMING,
                balance_after_transaction=receiver.balance.amount,
                intermediary_id=receiver.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=receiver.balance.id,
                    ) for conn in receiver.balance.connectors if conn.is_viewable()
                ],
            ),
        ]
    )
    session.add(transaction)

    session.execute(stmt, {'contract_id': connector.contract_id})

    session.commit()

    return transaction.to_dict(filter_for=account_id)


# Transaction


def create_funding(
        session: orm.Session,
        account_id: str,
        new_funding: api.Funding,
):
    _check_account(session, account_id)

    stmt = select(db.BalanceConnector)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.RECEIVABLE.value) > 0)\
        .limit(1)\
        .distinct()

    if new_funding.balance_selector:
        selector = new_funding.balance_selector

        if selector.type == api.BalanceSelectorType.BALANCE:
            stmt = stmt.where(db.Balance.id == selector.id)
        elif selector.type == api.BalanceSelectorType.CONNECTOR:
            stmt = stmt.where(
                db.BalanceConnector.id == selector.id
            )
        else:
            raise Exception('Invalid selector')  # Never reach here.

    try:
        connector = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(
            status_code=422, detail='Available balance not found'
        )

    if connector.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")

    if connector.balance.amount + new_funding.amount > api.MAX_BALANCE:
        raise HTTPException(status_code=400, detail='Too large amount')

    connector.balance.amount += new_funding.amount

    transaction = db.Transaction(
        id=new_funding.id,
        kind=new_funding.kind,
        note=new_funding.attachment,
        intermediary_id=connector.intermediary_id,
        balance_updates=[
            db.BalanceUpdate(
                balance_id=connector.balance.id,
                amount=new_funding.amount,
                type=api.BalanceUpdateType.INCOMING,
                balance_after_transaction=connector.balance.amount,
                intermediary_id=connector.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_funding.id,
                        balance_id=connector.balance.id
                    ) for conn in connector.balance.connectors if conn.is_viewable()
                ],
            )
        ]
    )

    session.add(transaction)
    session.commit()

    return transaction.to_dict(filter_for=account_id)


def create_defunding(
        session: orm.Session,
        account_id: str,
        new_defunding: api.Defunding,
):
    _check_account(session, account_id)

    stmt = select(db.BalanceConnector)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.account_id == account_id)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.SPENDABLE.value) > 0)\
        .limit(1)\
        .distinct()

    if new_defunding.balance_selector:
        selector = new_defunding.balance_selector

        if selector.type == api.BalanceSelectorType.BALANCE:
            stmt = stmt.where(db.Balance.id == selector.id)
        elif selector.type == api.BalanceSelectorType.CONNECTOR:
            stmt = stmt.where(db.BalanceConnector.id == selector.id)
        else:
            raise Exception('Invalid selector')  # Never reach here.

    try:
        connector = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(
            status_code=422, detail='Available balance not found'
        )

    if connector.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")

    if connector.balance.amount < new_defunding.amount:
        raise HTTPException(
            status_code=400, detail='Insufficient balance amount')

    connector.balance.amount -= new_defunding.amount

    transaction = db.Transaction(
        id=new_defunding.id,
        kind=new_defunding.kind,
        note=new_defunding.attachment,
        intermediary_id=connector.intermediary_id,
        balance_updates=[
            db.BalanceUpdate(
                balance_id=connector.balance.id,
                amount=new_defunding.amount,
                type=api.BalanceUpdateType.OUTGOING,
                balance_after_transaction=connector.balance.amount,
                intermediary_id=connector.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_defunding.id,
                        balance_id=connector.balance.id
                    ) for conn in connector.balance.connectors if conn.is_viewable()
                ],
            )
        ]
    )

    session.add(transaction)
    session.commit()

    return transaction.to_dict(filter_for=account_id)


def create_payment(
        session: orm.Session,
        account_id: str,
        new_payment: api.Payment,
):
    _check_account(session, account_id)

    if account_id == new_payment.to:
        raise HTTPException(status_code=400, detail='Invalid payee specified')

    try:
        stmt = select(db.BalanceConnector)\
            .join(db.BalanceConnector.balance)\
            .where(db.BalanceConnector.permission.bitwise_and(db.Permission.RECEIVABLE.value) > 0)\
            .limit(1)\
            .distinct()

        if type(new_payment.to) == str:
            stmt = stmt.where(db.BalanceConnector.account_id == new_payment.to)
        elif new_payment.to.type == api.BalanceSelectorType.CONNECTOR:
            stmt = stmt.where(db.BalanceConnector.id == new_payment.to.id)
        else:
            raise Exception('Invalid selector')  # Never reach here.

        receiver_connector = session.scalars(stmt).one()

        stmt = select(db.BalanceConnector)\
            .join(db.BalanceConnector.balance)\
            .where(db.BalanceConnector.permission.bitwise_and(db.Permission.SPENDABLE.value) > 0)\
            .where(db.BalanceConnector.account_id == account_id)\
            .order_by(db.Balance.amount.desc())\
            .limit(1)

        if new_payment.source_balance_selector:
            selector = new_payment.source_balance_selector

            if selector.type == api.BalanceSelectorType.BALANCE:
                stmt = stmt.where(db.Balance.id == selector.id)
            elif selector.type == api.BalanceSelectorType.CONNECTOR:
                stmt = stmt.where(db.BalanceConnector.id == selector.id)
            else:
                raise Exception('Invalid selector')  # Never reach here.

        sender_connector = session.scalars(stmt).one()
    except NoResultFound:
        raise HTTPException(
            status_code=422, detail='Available balance not found'
        )

    if sender_connector.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")
    if receiver_connector.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")
    if sender_connector.balance.amount < new_payment.amount:
        raise HTTPException(status_code=400, detail='Insufficient balance')
    if receiver_connector.balance.amount + new_payment.amount > api.MAX_BALANCE:
        raise HTTPException(status_code=400, detail='Too large amount')

    sender_connector.balance.amount -= new_payment.amount
    receiver_connector.balance.amount += new_payment.amount

    transaction = db.Transaction(
        id=new_payment.id,
        kind=new_payment.kind,
        note=new_payment.attachment,
        intermediary_id=sender_connector.intermediary_id,
        balance_updates=[
            db.BalanceUpdate(
                balance_id=sender_connector.balance_id,
                amount=new_payment.amount,
                type=api.BalanceUpdateType.OUTGOING,
                balance_after_transaction=sender_connector.balance.amount,
                intermediary_id=sender_connector.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_payment.id,
                        balance_id=sender_connector.balance_id
                    ) for conn in sender_connector.balance.connectors if conn.is_viewable()
                ],
            ),
            db.BalanceUpdate(
                balance_id=receiver_connector.balance_id,
                amount=new_payment.amount,
                type=api.BalanceUpdateType.INCOMING,
                balance_after_transaction=receiver_connector.balance.amount,
                intermediary_id=receiver_connector.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_payment.id,
                        balance_id=receiver_connector.balance_id
                    ) for conn in receiver_connector.balance.connectors if conn.is_viewable()
                ],
            ),
        ]
    )

    session.add(transaction)
    session.commit()

    return transaction.to_dict(filter_for=account_id)


def create_transaction(
        session: orm.Session,
        new_transaction: api.CustomTransaction,
):
    transaction = db.Transaction(
        id=new_transaction.id,
        datetime=new_transaction.datetime,
        kind=new_transaction.kind,
        note=new_transaction.attachment,
        intermediary_id=new_transaction.intermediary_id,
        balance_updates=[],  # Update this in the following.
    )

    for bu in new_transaction.balance_updates:
        stmt = select(db.BalanceConnector)\
            .join(db.BalanceConnector.balance)\
            .where(db.BalanceConnector.id == bu.connector_id)\
            .limit(1)\
            .distinct()

        try:
            connector = session.scalars(stmt).one()
        except NoResultFound:
            raise HTTPException(
                status_code=400, detail='Invalid balance specified'
            )

        if bu.type == api.BalanceUpdateType.INCOMING:
            if connector.balance.amount + bu.amount > api.MAX_BALANCE:
                raise HTTPException(status_code=400, detail='Too large amount')
            connector.balance.amount += bu.amount
        elif bu.type == api.BalanceUpdateType.OUTGOING:
            if connector.balance.amount < bu.amount:
                raise HTTPException(
                    status_code=400, detail='Insufficient balance'
                )
            connector.balance.amount -= bu.amount
        else:
            # Never reach here.
            raise Exception(f'Invalid balance update type: {bu.type}')

        transaction.balance_updates.append(
            db.BalanceUpdate(
                transaction_id=new_transaction.id,
                balance_id=connector.balance_id,
                amount=bu.amount,
                type=bu.type,
                balance_after_transaction=connector.balance.amount,
                intermediary_id=connector.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=connector.balance_id,
                    ) for conn in connector.balance.connectors if conn.is_viewable()
                ],
            )
        )

    try:
        session.add(transaction)
        session.commit()
    except IntegrityError as e:
        raise HTTPException(
            status_code=400, detail=f'Integrity error: {str(e)}'
        )

    return transaction.to_dict()


def create_htlc(
        session: orm.Session,
        account_id: str,
        new_htlc: api.HTLC,
):
    _check_account(session, account_id)

    stmt = select(db.BalanceConnector)\
        .join(db.BalanceConnector.balance)\
        .where(db.BalanceConnector.permission.bitwise_and(db.Permission.SPENDABLE.value) > 0)\
        .order_by(db.Balance.amount.desc())\
        .limit(1)

    try:
        sender = session.scalars(
            stmt.where(db.BalanceConnector.account_id == account_id)
        ).one()
        receiver = session.scalars(
            stmt.where(db.BalanceConnector.account_id == new_htlc.to)
        ).one()
    except NoResultFound:
        raise HTTPException(
            status_code=400, detail='Cannot determine target balance'
        )

    # Prepare parameters.
    new_contract = api.HTLCContract(
        hash=new_htlc.hash,
        expires_at=new_htlc.expires_at,
    )
    new_balance = api.Balance(intermediary_id=sender.intermediary_id)
    new_sender_connector = api.HTLCConnector(
        intermediary_id=sender.intermediary_id,
        account_id=sender.account_id,
        balance_id=new_balance.id,
        kind=api.BalanceConnectorKind.HTLC_SENDING,
        contract_id=new_contract.id,
        permission=['Viewable'],
    )
    new_receiver_connector = api.HTLCConnector(
        intermediary_id=receiver.intermediary_id,
        account_id=receiver.account_id,
        balance_id=new_balance.id,
        kind=api.BalanceConnectorKind.HTLC_RECEIVING,
        contract_id=new_contract.id,
        permission=['Viewable'],
    )
    new_transaction = api.TransactionBase(kind=api.TransactionKind.TRANSFER)

    # Move amount to new balance for reserving.
    if sender.balance.state == api.BalanceState.INACTIVE:
        raise HTTPException(status_code=403, detail="Balance is inactive")

    if sender.balance.amount < new_htlc.amount:
        raise HTTPException(status_code=400, detail='Insufficient balance')

    sender.balance.amount -= new_htlc.amount
    new_balance = db.Balance(
        id=new_balance.id,
        state=new_balance.state,
        kind=new_balance.kind,
        amount=new_htlc.amount,
        intermediary_id=new_balance.intermediary_id,
    )

    # Make connectors for a sender and a receiver.
    sender_connector = db.HTLCConnector(
        id=new_sender_connector.id,
        inner=db.BalanceConnector(
            id=new_sender_connector.id,
            intermediary_id=new_sender_connector.intermediary_id,
            account_id=new_sender_connector.account_id,
            balance_id=new_sender_connector.balance_id,
            kind=new_sender_connector.kind,
            balance=new_balance,
            permission=db.Permission.from_list(
                new_sender_connector.permission
            ).value,
        ),
        contract_id=new_contract.id,
        contract=db.HTLC(
            id=new_contract.id,
            hash=new_contract.hash,
            expires_at=new_contract.expires_at,
        ),
    )
    receiver_connector = db.HTLCConnector(
        id=new_receiver_connector.id,
        inner=db.BalanceConnector(
            id=new_receiver_connector.id,
            intermediary_id=new_receiver_connector.intermediary_id,
            account_id=new_receiver_connector.account_id,
            balance_id=new_receiver_connector.balance_id,
            kind=new_receiver_connector.kind,
            permission=db.Permission.from_list(
                new_receiver_connector.permission
            ).value,
        ),
        contract_id=new_contract.id,
    )
    # Record transaction.
    transaction = db.Transaction(
        id=new_transaction.id,
        kind=new_transaction.kind,
        note=new_transaction.attachment,
        intermediary_id=sender.intermediary_id,
        balance_updates=[
            db.BalanceUpdate(
                balance_id=sender.balance.id,
                amount=new_htlc.amount,
                type=api.BalanceUpdateType.OUTGOING,
                balance_after_transaction=sender.balance.amount,
                intermediary_id=sender.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=conn.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=sender.balance.id,
                    ) for conn in sender.balance.connectors if conn.is_viewable()
                ],
            ),
            db.BalanceUpdate(
                balance_id=new_balance.id,
                amount=new_htlc.amount,
                type=api.BalanceUpdateType.INCOMING,
                balance_after_transaction=new_balance.amount,
                intermediary_id=sender.intermediary_id,
                available_for=[
                    db.BalanceHistory(
                        account_id=sender.account.id,
                        transaction_id=new_transaction.id,
                        balance_id=new_balance.id
                    ),
                    db.BalanceHistory(
                        account_id=receiver_connector.inner.account_id,
                        transaction_id=new_transaction.id,
                        balance_id=new_balance.id
                    ),
                ],
            ),
        ]
    )
    session.add_all([sender_connector, receiver_connector, transaction])
    session.commit()

    return {
        'contract': sender_connector.contract.to_dict(),
        'connectors': {
            'sender': sender_connector.inner.to_dict(),
            'receiver': receiver_connector.inner.to_dict(),
        },
        'transaction': transaction.to_dict(filter_for=account_id),
    }


def read_transactions(
        session: orm.Session,
        intermediary_id: str,
        kind: Optional[str] = None,
):
    stmt = select(db.Transaction)\
        .join(db.Transaction.balance_updates)\
        .where(db.Transaction.intermediary_id == intermediary_id)\
        .order_by(db.Transaction.datetime.desc())\
        .distinct()

    if kind:
        stmt = stmt.where(db.Transaction.kind == kind)

    transactions = session.scalars(stmt).all()

    return [tx.to_dict() for tx in transactions]


def read_transaction_by_id(
        session: orm.Session,
        transaction_id: str,
        account_id: Optional[str] = None,
        kind: Optional[str] = None,
):
    _check_if_account_exists(session, account_id)

    stmt = select(db.Transaction)\
        .join(db.Transaction.balance_updates)\
        .where(db.Transaction.id == transaction_id)\
        .distinct()

    if account_id:
        stmt = stmt\
            .join(db.BalanceUpdate.available_for)\
            .where(db.BalanceHistory.account_id == account_id)
    if kind:
        stmt = stmt.where(db.Transaction.kind == kind)

    try:
        return session.scalars(stmt).one().to_dict(filter_for=account_id)
    except NoResultFound:
        raise HTTPException(status_code=404, detail='No such transaction')


def read_transactions_on_account(
        session: orm.Session,
        account_id: str,
        kind: Optional[str] = None,
):
    _check_if_account_exists(session, account_id)

    stmt = select(db.Transaction)\
        .join(db.Transaction.balance_updates)\
        .join(db.BalanceUpdate.available_for)\
        .where(db.BalanceHistory.account_id == account_id)\
        .order_by(db.Transaction.datetime.desc())\
        .distinct()

    if kind:
        stmt = stmt.where(db.Transaction.kind == kind)

    transactions = session.scalars(stmt).all()

    return [tx.to_dict(filter_for=account_id) for tx in transactions]


def delete_transactions(
        session: orm.Session,
        intermediary_id: str,
        kind: Optional[str],
):
    stmt = delete(db.Transaction)\
        .where(db.Transaction.intermediary_id == intermediary_id)

    if kind:
        stmt = stmt.where(db.Transaction.kind == kind)

    result = session.execute(stmt)

    session.commit()

    return result.rowcount


def delete_transaction_by_id(session: orm.Session, transaction_id: str):
    result = session.execute(
        delete(db.Transaction).where(db.Transaction.id == transaction_id)
    )

    if result.rowcount == 0:
        raise HTTPException(status_code=404, detail='No such transaction')

    session.commit()
