import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.dialects.mysql import BIGINT, VARCHAR, CHAR, BOOLEAN, TIMESTAMP, TINYINT
from sqlalchemy.sql.functions import current_timestamp
from sqlalchemy.types import Integer

from typing import List

import enum


class Base(orm.DeclarativeBase):
    pass


class Intermediary(Base):
    __tablename__ = 'intermediary'

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    name = orm.mapped_column(VARCHAR(length=255), nullable=False)

    def to_dict(self):
        return {
            'id': self.id,
            'name': self.name,
        }


class Account(Base):
    __tablename__ = 'account'
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ['intermediary_id'],
            ['intermediary.id'],
            onupdate='CASCADE',
            ondelete='CASCADE',
        ),
    )

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    holder_name = orm.mapped_column(VARCHAR(length=255), nullable=False)
    state = orm.mapped_column(CHAR(length=16), nullable=False)
    kind = orm.mapped_column(CHAR(length=16), nullable=False)
    note = orm.mapped_column(VARCHAR(length=255), nullable=True)

    intermediary_id = orm.mapped_column(VARCHAR(length=255))

    connectors: orm.Mapped[List['BalanceConnector']] \
        = orm.relationship(back_populates='account')
    history: orm.Mapped[List['BalanceHistory']] = \
        orm.relationship(back_populates='account')

    def to_dict(self):
        return {
            'id': self.id,
            'intermediaryId': self.intermediary_id,
            'holderName': self.holder_name,
            'state': self.state,
            # 'note': self.note,
            'kind': self.kind,
        }


class Balance(Base):
    __tablename__ = 'balance'
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ['intermediary_id'],
            ['intermediary.id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    )

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    state = orm.mapped_column(CHAR(length=16), nullable=False)
    kind = orm.mapped_column(CHAR(length=16), nullable=False)
    amount = orm.mapped_column(BIGINT(unsigned=True), nullable=False)
    is_asset = orm.mapped_column(BOOLEAN, nullable=False, insert_default=True)

    intermediary_id = orm.mapped_column(VARCHAR(length=255))

    connectors: orm.Mapped[List['BalanceConnector']] \
        = orm.relationship(back_populates='balance')

    def to_dict(self, viewable=True):
        return {
            'id': self.id,
            # 'intermediaryId': self.intermediary_id,
            'state': self.state,
            'kind': self.kind,
            'amount': self.amount,
            # 'isAsset': self.is_asset,
        } if viewable else {
            'id': self.id,
            # 'intermediaryId': self.intermediary_id,
            'state': self.state,
            'kind': self.kind,
            # 'amount': self.amount,
            # 'isAsset': self.is_asset,
        }


    def to_dict_with_connector(self, connector):
        return {
            **self.to_dict(viewable=connector.is_viewable()),
            'connector': connector.to_compact_dict(),
        }


class Permission(enum.IntFlag):
    NONE = 0
    VIEWABLE = 1
    SPENDABLE = (1 << 1)
    RECEIVABLE = (1 << 2)

    @classmethod
    def to_list(cls, permission):
        ls = []
        if permission & cls.VIEWABLE:
            ls.append('Viewable')
        if permission & cls.SPENDABLE:
            ls.append('Spendable')
        if permission & cls.RECEIVABLE:
            ls.append('Receivable')
        return ls


    @classmethod
    def from_list(cls, ls):
        p = cls.NONE
        for item in ls:
            if item == 'Viewable':
                p |= cls.VIEWABLE
            if item == 'Spendable':
                p |= cls.SPENDABLE
            if item == 'Receivable':
                p |= cls.RECEIVABLE
        return p


class BalanceConnector(Base):
    __tablename__ = 'balance_connector'
    __table_args__ = (
        sa.UniqueConstraint('account_id', 'balance_id'),
        sa.ForeignKeyConstraint(
            ['account_id'], ['account.id'], onupdate="CASCADE", ondelete="CASCADE"),
        sa.ForeignKeyConstraint(
            ['balance_id'], ['balance.id'], onupdate="CASCADE", ondelete="CASCADE"),
        sa.ForeignKeyConstraint(
            ['intermediary_id'],
            ['intermediary.id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    )

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    kind = orm.mapped_column(CHAR(length=16), nullable=False)
    permission = orm.mapped_column(
        TINYINT(unsigned=True),
        nullable=False,
        default=(Permission.VIEWABLE | Permission.SPENDABLE | Permission.RECEIVABLE),
    )

    account_id = orm.mapped_column(VARCHAR(length=255))
    balance_id = orm.mapped_column(VARCHAR(length=255))
    intermediary_id = orm.mapped_column(VARCHAR(length=255))

    account: orm.Mapped['Account'] = \
        orm.relationship(back_populates="connectors")
    balance: orm.Mapped['Balance'] = \
        orm.relationship(back_populates="connectors")


    def to_dict(self):
        return {
            'id': self.id,
            'accountId': self.account_id,
            'balanceId': self.balance_id,
            # 'intermediaryId': self.intermediary_id,
            'kind': self.kind,
            'permission': Permission.to_list(self.permission),
        }


    def to_compact_dict(self):
        return {
            'id': self.id,
            'kind': self.kind,
            'permission': Permission.to_list(self.permission),
        }


    def is_viewable(self):
        return self.permission & Permission.VIEWABLE > 0


class Transaction(Base):
    __tablename__ = 'transaction'
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ['intermediary_id'],
            ['intermediary.id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    )

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    sub_id = orm.mapped_column(Integer)
    datetime = orm.mapped_column(
        TIMESTAMP, nullable=False, insert_default=current_timestamp(), index=True
    )
    kind = orm.mapped_column(CHAR(length=16), nullable=False)
    note = orm.mapped_column(VARCHAR(length=255), nullable=True)

    intermediary_id = orm.mapped_column(VARCHAR(length=255))

    balance_updates: orm.Mapped[List['BalanceUpdate']] = \
        orm.relationship(back_populates='transaction')

    def to_dict(self, filter_for=None):
        d = {
            'id': self.id,
            # 'subId': self.sub_id,
            'time': f'{self.datetime.isoformat(timespec="seconds")}Z',
            'kind': self.kind,
            'intermediaryId': self.intermediary_id,
            'balanceUpdates': [
                bu.to_dict(filter_for=filter_for) for bu in self.balance_updates
            ],
        }

        if self.note:
            d['attachment'] = self.note
        return d


class BalanceUpdate(Base):
    __tablename__ = 'balance_update'
    __table_args__ = (
        sa.PrimaryKeyConstraint('transaction_id', 'balance_id'),
        sa.ForeignKeyConstraint(
            ['transaction_id'],
            ['transaction.id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ['intermediary_id'],
            ['intermediary.id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    )

    # We don't set this ID a foreign key, because the target balance could be ephemeral.
    balance_id = orm.mapped_column(VARCHAR(length=255), nullable=False)
    amount = orm.mapped_column(BIGINT(unsigned=True), nullable=False)
    type = orm.mapped_column(CHAR(length=16), nullable=False)
    balance_after_transaction = \
        orm.mapped_column(BIGINT(unsigned=True), nullable=False)

    transaction_id = orm.mapped_column(VARCHAR(length=255))
    intermediary_id = orm.mapped_column(VARCHAR(length=255))

    transaction: orm.Mapped['Transaction'] = \
        orm.relationship(back_populates='balance_updates')
    available_for: orm.Mapped[List['BalanceHistory']] = \
        orm.relationship(back_populates='balance_update')

    def to_dict(self, filter_for=None):
        if filter_for and filter_for not in [h.account_id for h in self.available_for]:
            # Filter out others' sensitive information (balance).
            return {
                'balanceId': self.balance_id,
                'type': self.type,
                'amount': self.amount,
            }
        else:
            return {
                'balanceId': self.balance_id,
                'type': self.type,
                'amount': self.amount,
                'balanceAfterTransaction': self.balance_after_transaction,
            }


class BalanceHistory(Base):
    __tablename__ = 'balance_history'
    __table_args__ = (
        sa.PrimaryKeyConstraint('account_id', 'transaction_id', 'balance_id'),
        sa.ForeignKeyConstraint(
            ['account_id'],
            ['account.id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ['transaction_id', 'balance_id'],
            ['balance_update.transaction_id', 'balance_update.balance_id'],
            onupdate="CASCADE",
            ondelete="CASCADE",
        ),
    )

    account_id = orm.mapped_column(VARCHAR(length=255))
    transaction_id = orm.mapped_column(VARCHAR(length=255))
    balance_id = orm.mapped_column(VARCHAR(length=255))

    account: orm.Mapped['Account'] = orm.relationship(back_populates='history')
    balance_update: orm.Mapped['BalanceUpdate'] = orm.relationship()


class HTLC(Base):
    __tablename__ = 'htlc'

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    hash = orm.mapped_column(CHAR(length=64), nullable=False)
    expires_at = orm.mapped_column(TIMESTAMP, nullable=False, index=True)

    connectors: orm.Mapped[List['HTLCConnector']] = \
        orm.relationship(back_populates='contract')

    def to_dict(self):
        return {
            'id': self.id,
            'hash': self.hash,
            'expiresAt': f'{self.expires_at.isoformat(timespec="seconds")}Z',
        }


class HTLCConnector(Base):
    __tablename__ = 'htlc_connector'
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ['id'], ['balance_connector.id'], onupdate="CASCADE", ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ['contract_id'], ['htlc.id'], onupdate="CASCADE", ondelete="CASCADE",
        ),
    )

    id = orm.mapped_column(VARCHAR(length=255), primary_key=True)

    contract_id = orm.mapped_column(VARCHAR(length=255))

    inner: orm.Mapped['BalanceConnector'] = \
        orm.relationship()
    contract: orm.Mapped['HTLC'] = \
        orm.relationship(back_populates='connectors')

    def to_dict(self):
        return {
            **self.inner.to_dict(),
            'contract': self.contract.to_dict(),
        }
