import os
import traceback

from fastapi import Depends, FastAPI, Query, Header, Request
from fastapi.responses import JSONResponse

import mangum

from sqlalchemy.orm import Session

from typing import Optional

import crud
import api_schemas as api

# This import will cause the connection to the DB and related setup
# to be made.
import database


DEFAULT_LIMIT = 100


app = FastAPI()


#
# Utilities
#


def format_items(items, limit: int = DEFAULT_LIMIT, offset: int = 0):
    return {
        'total': len(items),
        # Quick implementation for controlling data volume of response.
        'results': items[offset:offset+limit],
    }


def format_exception(e):
    lines = list(traceback.TracebackException.from_exception(e).format())

    return lines[-1] + "\n" + "".join(lines[:-1])


@app.exception_handler(Exception)
async def exception_handler(_: Request, e: Exception):
    return JSONResponse(
        status_code=500,
        # Expose detail of internal errors for experiment.
        content={'message': f'{type(e)}: {str(e)}'},
        # content={'message': format_exception(e)},
    )


#
# Dependencies
#


def get_db(req: Request):
    if 'aws.event' in req.scope:
        version = req.scope['aws.event']['stageVariables']['version']
    else:
        version = 'local'

    yield from database.get_db(version)


#
# API implementations
#


@app.get('/internal/accounts')
async def internal_get_accounts(
        intermediary_id: str = Query(alias='intermediaryId'),
        with_balance: str = Query(default='0', alias='withBalance'),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    if with_balance != '0':
        accounts = crud.read_accounts_with_balance(db, intermediary_id)

        return format_items(accounts, limit=limit, offset=offset)
    else:
        accounts = crud.read_accounts(db, intermediary_id)

        return format_items(crud.read_accounts(db, intermediary_id), limit=limit, offset=offset)


@app.post('/internal/accounts', status_code=201)
async def internal_make_account(
        new_account: api.Account,
        db: Session = Depends(get_db),
):
    return crud.create_account(db, new_account)


@app.get('/internal/accounts/{account_id}')
async def internal_get_account(
        account_id: str,
        with_balance: str = Query(default='0', alias='withBalance'),
        db: Session = Depends(get_db),
):
    if with_balance != '0':
        return crud.read_account_by_id_with_balance(db, account_id)
    else:
        return crud.read_account_by_id(db, account_id)


@app.patch('/internal/accounts/{account_id}')
async def internal_update_account(
        account_id: str,
        update: api.AccountUpdate,
        db: Session = Depends(get_db),
):
    return crud.update_account(db, account_id, update)


@app.delete('/internal/accounts/{account_id}', status_code=204)
async def internal_delete_account(
        account_id: str,
        db: Session = Depends(get_db),
):
    crud.delete_account_by_id(db, account_id)


@app.get('/internal/balances')
async def internal_get_balances(
        intermediary_id: str = Query(alias='intermediaryId'),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    balances = crud.read_balances(db, intermediary_id)

    return format_items(balances, limit=limit, offset=offset)


@app.post('/internal/balances', status_code=201)
async def internal_make_balance(
        new_balance: api.BalanceWithConnectors,
        db: Session = Depends(get_db),
):
    return crud.create_balance(db, new_balance)


@app.patch('/internal/balances/{balance_id}')
async def internal_update_balance(
        balance_id: str,
        update: api.BalanceStateUpdate,
        db: Session = Depends(get_db),
):
    return crud.update_balance(db, balance_id, update)


@app.get('/internal/balances/{balance_id}')
async def internal_get_balance(balance_id: str, db: Session = Depends(get_db)):
    return crud.read_balance_by_id(db, balance_id)


@app.delete('/internal/balances/{balance_id}', status_code=204)
async def internal_delete_balance(
        balance_id: str,
        db: Session = Depends(get_db),
):
    crud.delete_balance_by_id(db, balance_id)


@app.get('/internal/balance-connectors')
async def internal_get_balance_connectors(
        intermediary_id: str = Query(alias='intermediaryId'),
        account_id: Optional[str] = Query(alias='accountId', default=None),
        balance_id: Optional[str] = Query(alias='balanceId', default=None),
        kind: Optional[str] = Query(default=None),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    connectors = crud.read_connectors(
        db,
        intermediary_id,
        account_id=account_id,
        balance_id=balance_id,
        kind=kind
    )

    return format_items(connectors, limit=limit, offset=offset)


@app.post('/internal/balance-connectors', status_code=201)
async def internal_connect_balance_to_account(
        new_connector: api.BalanceConnector,
        db: Session = Depends(get_db),
):
    return crud.create_connector(db, new_connector)


@app.get('/internal/balance-connectors/{connector_id}')
async def internal_get_balance_connector(
        connector_id: str,
        db: Session = Depends(get_db),
):
    return crud.read_connector_by_id(db, connector_id)


@app.delete('/internal/balance-connectors/{connector_id}', status_code=204)
async def internal_delete_balance_connector(
        connector_id: str,
        db: Session = Depends(get_db),
):
    crud.delete_connector_by_id(db, connector_id)


@app.post('/internal/transactions')
async def internal_make_transaction(
        new_transaction: api.CustomTransaction,
        db: Session = Depends(get_db),
):
    return crud.create_transaction(db, new_transaction)


@app.get('/internal/transactions')
async def internal_get_transactions(
        intermediary_id: str = Query(alias='intermediaryId'),
        kind: Optional[str] = Query(default=None),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    transactions = crud.read_transactions(db, intermediary_id, kind)

    return format_items(transactions, limit=limit, offset=offset)


@app.delete('/internal/transactions', status_code=204)
async def internal_delete_transactions(
        intermediary_id: str = Query(alias='intermediaryId'),
        kind: Optional[str] = Query(default=None),
        db: Session = Depends(get_db),
):
    return crud.delete_transactions(db, intermediary_id, kind)


@app.get('/internal/transactions/{transaction_id}')
async def internal_get_transaction(
        transaction_id: str,
        db: Session = Depends(get_db),
):
    return crud.read_transaction_by_id(db, transaction_id)


@app.delete('/internal/transactions/{transaction_id}', status_code=204)
async def internal_delete_transaction(
        transaction_id: str,
        db: Session = Depends(get_db),
):
    return crud.delete_transaction_by_id(db, transaction_id)


# Basic APIs


@app.get('/version')
async def version(req: Request):
    if 'aws.event' in req.scope:
        return {
            'version': req.scope['aws.event']['stageVariables']['version'],
        }
    else:
        return {'version': 'local'}


@app.get('/intermediaries')
async def get_intermediaries(
        db: Session = Depends(get_db),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
):
    try:
        return format_items(crud.read_intermediaries(db), limit=limit, offset=offset)
    except Exception as e:
        return {'message': str(e)}


@app.get('/account')
async def get_account(
        account_id: str = Header(),
        with_balance: str = Query(default='0', alias='withBalance'),
        db: Session = Depends(get_db),
):
    if with_balance == '0':
        return crud.read_account_by_id(db, account_id)
    else:
        return crud.read_account_by_id_with_balance(db, account_id)


@app.get('/transactions')
async def get_transactions(
        account_id: str = Header(),
        kind: Optional[str] = Query(default=None),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    transactions = crud.read_transactions_on_account(db, account_id, kind)

    return format_items(transactions, limit=limit, offset=offset)


@app.get('/transactions/{transaction_id}')
async def get_transaction(
        transaction_id: str,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.read_transaction_by_id(db, transaction_id, account_id=account_id)


@app.get('/balances')
async def get_connected_balances(
        account_id: str = Header(),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    balances = crud.read_balances_connected_to_account(db, account_id)

    return format_items(balances, limit=limit, offset=offset)


@app.get('/balances/{balance_id}')
async def get_connected_balance(
        balance_id: str,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.read_balance_connected_to_account_by_id(
        db,
        account_id,
        balance_id,
    )


@app.get('/fundings')
async def get_fundings(
        account_id: str = Header(),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    transactions = crud.read_transactions_on_account(
        db,
        account_id,
        api.TransactionKind.FUNDING,
    )

    return format_items(transactions, limit=limit, offset=offset)


@app.post('/fundings')
async def fund(
        new_funding: api.Funding,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.create_funding(db, account_id, new_funding)


@app.get('/fundings/{transaction_id}')
async def get_funding(
        transaction_id: str,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.read_transaction_by_id(
        db,
        transaction_id,
        account_id=account_id,
        kind=api.TransactionKind.FUNDING,
    )


@app.get('/defundings')
async def get_defundings(
        account_id: str = Header(),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    transactions = crud.read_transactions_on_account(
        db,
        account_id,
        api.TransactionKind.DEFUNDING,
    )

    return format_items(transactions, limit=limit, offset=offset)


@app.post('/defundings')
async def defund(
        new_defunding: api.Defunding,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.create_defunding(db, account_id, new_defunding)


@app.get('/defundings/{transaction_id}')
async def get_defunding(
        transaction_id: str,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.read_transaction_by_id(
        db,
        transaction_id,
        account_id=account_id,
        kind=api.TransactionKind.DEFUNDING,
    )


@app.get('/payments')
async def get_payments(
        account_id: str = Header(),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
        db: Session = Depends(get_db),
):
    transactions = crud.read_transactions_on_account(
        db,
        account_id,
        api.TransactionKind.TRANSFER,
    )

    return format_items(transactions, limit=limit, offset=offset)


@app.post('/payments')
async def make_payment(
        new_payment: api.Payment,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.create_payment(db, account_id, new_payment)


@app.get('/payments/{transaction_id}')
async def get_payment(
        transaction_id: str,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.read_transaction_by_id(
        db,
        transaction_id,
        account_id=account_id,
        kind=api.TransactionKind.TRANSFER,
    )


@app.get('/hashed-timelocks')
def get_hashed_timelocks(
        account_id: str = Header(),
        db: Session = Depends(get_db),
        limit: int = Query(default=DEFAULT_LIMIT),
        offset: int = Query(default=0),
):
    connectors = crud.read_htlc_connectors(db, account_id)

    return format_items(connectors, limit=limit, offset=offset)


@app.post('/hashed-timelocks')
def make_hashed_timelock(
        new_htlc: api.HTLC,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.create_htlc(db, account_id, new_htlc)


@app.get('/hashed-timelocks/{connector_id}')
def get_hashed_timelock(connector_id: str, db: Session = Depends(get_db)):
    return crud.read_htlc_connector_by_id(db, connector_id)


@app.delete('/hashed-timelocks/{connector_id}')
def cancel_hashed_timelock(
        connector_id: str,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.delete_htlc_connector_with_retake(db, account_id, connector_id)


@app.put('/hashed-timelocks/{connector_id}/secret')
def release_hashed_timelock(
        connector_id: str,
        secret: api.HTLCSecret,
        account_id: str = Header(),
        db: Session = Depends(get_db),
):
    return crud.delete_htlc_connector_with_claim(
        db, account_id, connector_id, secret
    )


@app.post('/admin/init')
def reset_db(preset: api.Preset):
    database.reset_db(preset)

    return 'DB initalization completed'


handler = mangum.Mangum(app)
